/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * dsndbd.c
 *
 *  Created on: Feb 25, 2010
 *      Author: drewyrr
 */

#include "dsndbd.h"

#include <stdio.h>
#include <apr_lib.h>
#include <apr_buckets.h>

static dsnconn_rec *dsnconn_make(dsndbd_rec *dsndbd, apr_pool_t *parent) {
	apr_pool_t *pool;
	dsnconn_rec *con;

	apr_pool_create(&pool, parent);
	con = apr_pcalloc(pool, sizeof(dsnconn_rec));
	con->pool = pool;
	con->dsndbd = dsndbd;

	return con;
}

apr_status_t dsndbd_close_database(dsnconn_rec *con, dsndbd_rec *conf,
		apr_pool_t *pool) {
	apr_status_t status = apr_dbd_close(conf->driver, con->handle);

	apr_pool_destroy(con->pool);

	return status;
}

apr_status_t dsndbd_open_database(dsnconn_rec **pcon, dsndbd_rec *conf,
		apr_pool_t *pool) {
	const char *error = NULL;
	apr_status_t status;
	int errnum;
	dsnconn_rec *con = dsnconn_make(conf, pool);
	*pcon = con;

	// TODO make me configurable
	conf->max_body_size = HUGE_STRING_LEN;

	apr_dbd_init(pool);
	/* allocate the driver out of the dsndbd pool */
	switch (apr_dbd_get_driver(conf->pool, conf->driver_name, &conf->driver)) {
	case APR_ENOTIMPL:
		//		if (!quiet) {
		fprintf(stderr, "DBD: No driver for %s", conf->driver_name);
		//		}
		return APR_ENOTIMPL;
	case APR_EDSOOPEN:
		//		if (!quiet) {
		fprintf(stderr, "DBD: Can't load driver file apr_dbd_%s.so",
				conf->driver_name);
		//		}
		return APR_EDSOOPEN;
	case APR_ESYMNOTFOUND:
		//		if (!quiet) {
		fprintf(
				stderr,
				"DBD: Failed to load driver apr_dbd_%s_driver (symbol not found)",
				conf->driver_name);
		//		}
		return APR_ESYMNOTFOUND;
	}

	if (APR_SUCCESS != (status = apr_dbd_open_ex(conf->driver, con->pool,
			conf->params, &con->handle, &error))) {
		fprintf(
				stderr,
				"DBD: Failed to open a connection to the database (using %s): %s",
				conf->driver_name, error);
		return status;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, con->pool, con->handle,
			"select key from message where message_id = %s", "select-message",
			&con->select_message))) {
		fprintf(stderr,
				"DBD: Failed to prepare message statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"insert into message ('message_id') values (%s)", "insert-message",
			&con->insert_message))) {
		fprintf(stderr,
				"DBD: Failed to prepare message statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"select key from header where header = %s", "select-header",
			&con->select_header))) {
		fprintf(stderr,
				"DBD: Failed to prepare header statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"insert into header ('header') values (%s)", "insert-header",
			&con->insert_header))) {
		fprintf(stderr,
				"DBD: Failed to prepare header statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"select key from value where value = %s", "select-value",
			&con->select_value))) {
		fprintf(stderr,
				"DBD: Failed to prepare value statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"insert into value ('value') values (%s)", "insert-value",
			&con->insert_value))) {
		fprintf(stderr,
				"DBD: Failed to prepare value statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"select key from recipient where message_key = %llu and value_key = %llu",
			"select-message", &con->select_recipient))) {
		fprintf(stderr,
				"DBD: Failed to prepare recipient statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"insert into recipient ('message_key', 'value_key') values (%llu, %llu)",
					"insert-recipient", &con->insert_recipient))) {
		fprintf(stderr,
				"DBD: Failed to prepare recipient statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"select key from recipient_headers where recipient_key = %llu and header_key = %llu and value = %s",
					"select-recipient-headers", &con->select_recipient_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare recipient headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"insert into recipient_headers ('recipient_key', 'header_key', 'value') values (%llu, %llu, %s)",
					"insert-recipient-headers", &con->insert_recipient_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare recipient headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"select key from recipient_headers where recipient_key = %llu and header_key = %llu and value_key = %llu",
					"select-indexed-recipient-headers",
					&con->select_indexed_recipient_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare indexed recipient headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"insert into recipient_headers ('recipient_key', 'header_key', 'value_key') values (%llu, %llu, %llu)",
					"insert-indexed-recipient-headers",
					&con->insert_indexed_recipient_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare indexed recipient headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum = apr_dbd_prepare(conf->driver, conf->pool, con->handle,
			"select key from notification where message_id = %s and "
				"recipient_key = %llu and body = %s", "select-notification",
			&con->select_notification))) {
		fprintf(stderr,
				"DBD: Failed to prepare notification statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"insert into notification ('message_id', 'recipient_key', 'body') values (%s,%llu,%s)",
					"insert-notification", &con->insert_notification))) {
		fprintf(stderr,
				"DBD: Failed to prepare notification statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"select key from notification_headers where notification_key = %llu and header_key = %llu and value = %s",
					"select-notification-headers",
					&con->select_notification_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare notification headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"insert into notification_headers ('notification_key', 'header_key', 'value') values (%llu, %llu, %s)",
					"insert-notification-headers",
					&con->insert_notification_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare notification headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"select key from notification_headers where notification_key = %llu and header_key = %llu and value_key = %llu",
					"select-indexed-notification-headers",
					&con->select_indexed_notification_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare indexed notification headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	if ((errnum
			= apr_dbd_prepare(
					conf->driver,
					conf->pool,
					con->handle,
					"insert into notification_headers ('notification_key', 'header_key', 'value_key') values (%llu, %llu, %llu)",
					"insert-indexed-notification-headers",
					&con->insert_indexed_notification_headers))) {
		fprintf(
				stderr,
				"DBD: Failed to prepare indexed notification headers statement (using %s): %s",
				conf->driver_name, apr_dbd_error(conf->driver, con->handle,
						errnum));
		return APR_EGENERAL;
	}

	return APR_SUCCESS;
}

static apr_status_t update_kv(apr_pool_t *pool, dsnconn_rec *dsncon,
		const void **values, apr_uint64_t *key,
		apr_dbd_prepared_t *select_statement,
		apr_dbd_prepared_t *insert_statement) {
	int errnum;
	apr_dbd_results_t *res = NULL;
	apr_dbd_row_t *row = NULL;
	int written = 0;

	while (1) {
		if (!(errnum = apr_dbd_pbselect(dsncon->dsndbd->driver, pool,
				dsncon->handle, &res, select_statement, 0, values))) {
			if (!apr_dbd_get_row(dsncon->dsndbd->driver, pool, res, &row, -1)) {

				apr_dbd_datum_get(dsncon->dsndbd->driver, row, 0,
						APR_DBD_TYPE_ULONGLONG, key);

				return APR_SUCCESS;

			} else {
				int nrows = 0;

				/* only want to read? if so, leave now */
				if (!insert_statement) {
					return APR_SUCCESS;
				}

				if (written) {
					fprintf(
							stderr,
							"DBD: Sanity check failed, insert succeeded but select failed (using %s): %s",
							dsncon->dsndbd->driver_name, apr_dbd_error(
									dsncon->dsndbd->driver, dsncon->handle,
									errnum));
					return APR_EGENERAL;
				}

				if (!(errnum = apr_dbd_pbquery(dsncon->dsndbd->driver, pool,
						dsncon->handle, &nrows, insert_statement, values))) {

					/* we wrote the row, now do the select again */
					written = 1;
					continue;

				} else {
					fprintf(
							stderr,
							"DBD: Failed to run insert statement to insert message_id (using %s): %s",
							dsncon->dsndbd->driver_name, apr_dbd_error(
									dsncon->dsndbd->driver, dsncon->handle,
									errnum));
					return APR_EGENERAL;
				}

			}
		} else {
			fprintf(
					stderr,
					"DBD: Failed to run select statement to find message_id (using %s): %s",
					dsncon->dsndbd->driver_name, apr_dbd_error(
							dsncon->dsndbd->driver, dsncon->handle, errnum));
			return APR_EGENERAL;
		}
	}

}

typedef struct headers_t {
	apr_pool_t *pool;
	dsnconn_rec *dsncon;
	apr_status_t status;
	apr_uint64_t key;
	apr_dbd_prepared_t *select;
	apr_dbd_prepared_t *insert;
	apr_dbd_prepared_t *select_indexed;
	apr_dbd_prepared_t *insert_indexed;
} headers_t;

static int do_headers(void *dummy, const char *key, const char *value) {
	headers_t *hdrs = (headers_t *) dummy;
	dsnconn_rec *dsncon = hdrs->dsncon;
	apr_uint64_t header_key = 0, value_key = 0;
	int index = 0;

	if (key[0] == 'A') {
		if (!strcmp(key, "Action")) {
			index = 1;
		}
	} else if (key[0] == 'B') {
		if (!strcmp(key, "BCC")) {
			index = 1;
		}
	} else if (key[0] == 'C') {
		if (!strcmp(key, "CC")) {
			index = 1;
		}
	} else if (key[0] == 'S') {
		if (!strcmp(key, "Status")) {
			index = 1;
		}
	} else if (key[0] == 'T') {
		if (!strcmp(key, "To")) {
			index = 1;
		}
	}

	/* handle headers */
	{
		const void *values[1];
		values[0] = key;

		hdrs->status = update_kv(hdrs->pool, dsncon, values, &header_key,
				dsncon->select_header, dsncon->insert_header);
		if (APR_SUCCESS != hdrs->status) {
			return 0;
		}

	}

	/* handle indexed values */
	if (index) {
		const void *values[3];
		values[0] = value;

		hdrs->status = update_kv(hdrs->pool, dsncon, values, &value_key,
				dsncon->select_value, dsncon->insert_value);
		if (APR_SUCCESS != hdrs->status) {
			return 0;
		}

		values[0] = &hdrs->key;
		values[1] = &header_key;
		values[2] = &value_key;

		hdrs->status = update_kv(hdrs->pool, dsncon, values, &header_key,
				hdrs->select_indexed, hdrs->insert_indexed);
		if (APR_SUCCESS != hdrs->status) {
			return 0;
		}

	}

	/* try write the row */
	else {
		const void *values[3];
		values[0] = &hdrs->key;
		values[1] = &header_key;
		values[2] = value;

		hdrs->status = update_kv(hdrs->pool, dsncon, values, &header_key,
				hdrs->select, hdrs->insert);
		if (APR_SUCCESS != hdrs->status) {
			return 0;
		}

	}

	return 1;
}

apr_status_t dsndbd_update_message(apr_pool_t *pool, dsnconn_rec *dsncon,
		const char *recipient, apr_table_t *headers, apr_uint64_t *key) {
	const char *message_id = apr_table_get(headers, "Message-Id");
	apr_status_t status;
	apr_uint64_t message_key = 0, value_key = 0, recipient_key = 0;

	if (message_id) {
		const void *values[1];
		values[0] = message_id;

		status = update_kv(pool, dsncon, values, &message_key,
				dsncon->select_message, dsncon->insert_message);
		if (APR_SUCCESS != status) {
			return status;
		}

	} else {
		fprintf(stderr, "DSN: Message lacked a Message-ID\n");
		return APR_EGENERAL;
	}

	if (recipient) {
		const void *values[1];
		values[0] = recipient;

		status = update_kv(pool, dsncon, values, &value_key,
				dsncon->select_value, dsncon->insert_value);
		if (APR_SUCCESS != status) {
			return status;
		}

	} else {
		fprintf(stderr, "DSN: Message lacked a VERP enabled To address\n");
		return APR_EGENERAL;
	}

	if (message_key && value_key) {
		const void *values[2];
		values[0] = &message_key;
		values[1] = &value_key;

		status = update_kv(pool, dsncon, values, &recipient_key,
				dsncon->select_recipient, dsncon->insert_recipient);
		if (APR_SUCCESS != status) {
			return status;
		}

		*key = recipient_key;

	} else {
		fprintf(
				stderr,
				"DSN: Sanity check failed: message_key %" APR_UINT64_T_FMT " / value key %" APR_UINT64_T_FMT "\n",
				message_key, value_key);
		return APR_EGENERAL;
	}

	/* handle the recipient headers */
	{
		headers_t hdrs;
		hdrs.pool = pool;
		hdrs.dsncon = dsncon;
		hdrs.status = APR_SUCCESS;
		hdrs.key = recipient_key;
		hdrs.select = dsncon->select_recipient_headers;
		hdrs.insert = dsncon->insert_recipient_headers;
		hdrs.select_indexed = dsncon->select_indexed_recipient_headers;
		hdrs.insert_indexed = dsncon->insert_indexed_recipient_headers;

		apr_table_do(do_headers, &hdrs, headers, "To", "CC", "Subject", "Received", NULL);

		if (APR_SUCCESS == hdrs.status) {
			return hdrs.status;
		}
	}

	return APR_SUCCESS;
}

apr_status_t dsndbd_update_notification(apr_pool_t *pool, dsnconn_rec *dsncon,
		const char *message_id, apr_uint64_t recipient_key,
		apr_table_t *headers, apr_table_t *headers_body,
		apr_bucket_brigade *body) {
	apr_uint64_t notification_key = 0;

	if (!message_id) {
		fprintf(stderr, "DSN: Notification lacked a Message-ID\n");
		return APR_EGENERAL;
	}

	/* handle the notification */
	{
		char *buf = apr_pcalloc(pool, dsncon->dsndbd->max_body_size + 1);
		apr_size_t len = dsncon->dsndbd->max_body_size;
		apr_status_t status = APR_SUCCESS;

		apr_brigade_flatten(body, buf, &len);
		if (len > 0) {
			const void *values[3];
			values[0] = message_id;
			values[1] = &recipient_key;
			values[2] = buf;

			status = update_kv(pool, dsncon, values, &notification_key,
					dsncon->select_notification, dsncon->insert_notification);
			if (APR_SUCCESS != status) {
				return status;
			}
		}
	}

	/* handle the notification headers */
	{
		headers_t hdrs;
		hdrs.pool = pool;
		hdrs.dsncon = dsncon;
		hdrs.status = APR_SUCCESS;
		hdrs.key = notification_key;
		hdrs.select = dsncon->select_notification_headers;
		hdrs.insert = dsncon->insert_notification_headers;
		hdrs.select_indexed = dsncon->select_indexed_notification_headers;
		hdrs.insert_indexed = dsncon->insert_indexed_notification_headers;

		apr_table_do(do_headers, &hdrs, headers, "Reporting-MTA",
				"Arrival-Date", "Final-Recipient", "Original-Recipient",
				"Action", "Status", "Diagnostic-Code", NULL);
		if (hdrs.status != APR_SUCCESS) {
			return hdrs.status;
		}

		apr_table_do(do_headers, &hdrs, headers_body, "Reporting-MTA",
				"Arrival-Date", "Final-Recipient", "Original-Recipient",
				"Action", "Status", "Diagnostic-Code", NULL);

		if (hdrs.status != APR_SUCCESS) {
			return hdrs.status;
		}
	}

	return APR_SUCCESS;
}

dsndbd_rec *dsndbd_make(apr_pool_t *parent) {
	apr_pool_t *pool;
	dsndbd_rec *dsndbd;

	apr_pool_create(&pool, parent);
	dsndbd = apr_pcalloc(pool, sizeof(dsndbd_rec));
	dsndbd->pool = pool;

	apr_pool_userdata_setn(dsndbd, "dsndbd", NULL, parent);

	return dsndbd;
}
