/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * processdsn.c
 *
 * Parse delivery status notifications, write the results to a SQL database.
 *
 * The database is optimised for data size at the cost of speed, as we expect
 * to handle large volumes of data. We have the luxury of time, as the data
 * processing happens after emails have been queued for delivery.
 *
 * For each delivery status notification, extract the following information:
 *
 * - The following headers from the original message, saved to a message table:
 *   o Received
 *   o Date
 *   o Subject
 *   o To
 *   o From
 *   o Message-ID
 *
 * The message is created if it doesn't exist, keyed by Message-ID. If the
 * original message is missing, ignore the notification. Original message is
 * in message/rfc822.
 *
 * - The following headers from the DSN, saved to the notification table:
 *   o Reporting-MTA
 *   o Arrival-Date
 *   o Final-Recipient
 *   o Original-Recipient
 *   o Action
 *   o Status
 *   o Diagnostic-Code
 *
 * The notification is created if it doesn't already exist, keyed by notification
 * Message-ID. The notification is in message/delivery-status.
 *
 * - Optionally store the human readable error message in the error table.
 *
 * The error message is in any section marked text/ *. If there are more than one
 * messages, store all of them.
 */

#include <apr_strings.h>

#include <apr.h>
#include <apr_lib.h>
#include <apu_version.h>
#include <apu.h>
#include <apr_getopt.h>
#include <apr_tables.h>
#include <apr_dbd.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "config.h"

#include "mimeparse.h"
#include "dsndbd.h"
#include "dsn.h"

#define HELP_HEADER "Usage: %s [--driver dbdriver] [--params dbparams] [--version] [--help]\n" \
    "\n" \
    "Process a delivery status notification, and write the results to a SQL database.\n"
#define HELP_FOOTER "\nNote: in some cases, database passwords might be required within \n" \
    "the parameter given. As an alternative to passing the parameter on\n" \
    "the command line, the " DBD_DRIVER " and " DBD_PARAMS " environment\n" \
    "variables can be used to pass the database parameters.\n" \
    "\nReport bugs to " PACKAGE_BUGREPORT "."
#define MAX_BUFFER_SIZE 8192
#define DBD_DRIVER "DBD_DRIVER"
#define DBD_PARAMS "DBD_PARAMS"
#define EOL "\n"

static int quiet = 0;

apr_status_t read_pipe(mimeparse_rec *conf) {

	apr_status_t status;
	apr_size_t len = 0;
	char errbuf[MAX_BUFFER_SIZE];
	mime_rec *mime = mime_make(conf->pool, conf, NULL);

	apr_bucket *b;
	apr_bucket_brigade *bb, *obb;

	bb = apr_brigade_create(conf->pool, conf->alloc);
	obb = apr_brigade_create(conf->pool, conf->alloc);
	b = apr_bucket_pipe_create(conf->in, conf->alloc);

	APR_BRIGADE_INSERT_HEAD(bb, b);

	do {

		status = apr_brigade_split_line(obb, bb, APR_BLOCK_READ,
				HUGE_STRING_LEN);

		if (APR_SUCCESS == status || APR_EOF == status) {
			char str[HUGE_STRING_LEN];
			len = HUGE_STRING_LEN;

			apr_brigade_flatten(obb, str, &len);

			if (len > 0) {
				status = mime_process_line(conf, mime, str, len);
				if (APR_SUCCESS != status) {
					apr_file_printf(conf->err,
							"DBD: Could not process line: %s: %s\n",
							apr_strerror(status, errbuf, sizeof(errbuf)),
							apr_pstrndup(conf->pool, str, len));
					break;
				}
			}

			/* write to output file */
			if (conf->out && len > 0) {
				status = apr_file_write_full(conf->out, str, len, &len);
				if (APR_SUCCESS != status) {
					apr_file_printf(conf->err,
							"DBD: Could not write to output file: %s\n",
							apr_strerror(status, errbuf, sizeof(errbuf)));
					break;
				}
			}

			apr_brigade_cleanup(obb);
		} else {
			apr_file_printf(conf->err,
					"DBD: Could not read from input file: %s\n", apr_strerror(
							status, errbuf, sizeof(errbuf)));
			break;
		}

		if (len <= 0) {
			status = mime_last_line(conf, mime);
			if (APR_SUCCESS != status) {
				apr_file_printf(conf->err,
						"DBD: Could not process last line: %s\n", apr_strerror(
								status, errbuf, sizeof(errbuf)));
			}
			break;
		}

	} while (1);

	return status;
}

/**
 * Who are we again?
 */
static void version(const char * const argv[]) {
	printf("%s (%s)\n", argv[0], PACKAGE_STRING);
}

/**
 * Help the long suffering end user.
 */
static void help(const char * header, const char *footer,
		const apr_getopt_option_t *opts) {
	int i = 0;

	if (header) {
		printf("%s\n", header);
	}

	for (i = 0; opts[i].description; i++) {
		printf("%s\n", opts[i].description);
	}

	if (footer) {
		printf("%s\n", footer);
	}
}

int main(int argc, const char * const argv[]) {
	dsn_rec *dsn;
	dsndbd_rec *dsndbd;
	dsnconn_rec *dsncon;
	apr_pool_t *pmain = NULL;
	apr_getopt_t *opt;
	const char *optarg;
	apr_status_t status = APR_SUCCESS;
	int optch;
	int errno;
	char errbuf[MAX_BUFFER_SIZE];
	static const apr_getopt_option_t
			opts[] = {
					{ "help", 'h', 0, " --help\t\t\t\tDisplay this help." },
					{ "version", 'v', 0, " --version\t\t\tDisplay the version." },
					{ "quiet", 'q', 0, " -q, --quiet\t\t\tBe quiet." },
					{ "file", 'f', 1,
							" -f, --file=file\t\tFile to read DSN from. Defaults to stdin." },
					{ "out", 'o', 1,
							" -o, --out=file\t\t\tOptional file to write DSN to." },
					{ "driver", 'd', 1,
							" -d, --driver=driver\t\tDatabase driver to use." },
					{ "params", 'p', 1,
							" -p, --params=params\t\tDatabase parameter string." },
					{ "params-file", 'c', 1,
							" -c, --params-file=file\t\tFile containing database parameter string." },
					{ NULL, 0, 0, NULL } };

	/* lets get APR off the ground, and make sure it terminates cleanly */
	if (APR_SUCCESS != (status = apr_app_initialize(&argc, &argv, NULL))) {
		fprintf(stderr, "could not apr_app_initialize(). Oops.");
		exit(1);
	}
	atexit(apr_terminate);
	if (APR_SUCCESS != (status = apr_pool_create(&pmain, NULL))) {
		fprintf(stderr, "could not create the root memory pool. Oops.");
		exit(1);
	}

	/* initial setup of streams */
	dsn = dsn_make(pmain);
	dsndbd = dsndbd_make(pmain);

	dsndbd->driver_name = getenv(DBD_DRIVER);
	dsndbd->params = getenv(DBD_PARAMS);

	/* initial scan of global command line options */
	apr_getopt_init(&opt, pmain, argc, argv);
	while ((status = apr_getopt_long(opt, opts, &optch, &optarg))
			== APR_SUCCESS) {

		switch (optch) {
		case 'q': {
			quiet++;
			break;
		}
		case 'v': {
			version(argv);
			exit(0);
		}
		case 'h': {
			help((const char *) apr_psprintf(pmain, HELP_HEADER, argv[0]),
					HELP_FOOTER, opts);
			exit(0);
		}
		case 'd': {
			dsndbd->driver_name = optarg;
			break;
		}
		case 'p': {
			dsndbd->params = optarg;
			break;
		}
		case 'c': {
			apr_file_t *config;
			apr_size_t len;
			char buffer[MAX_BUFFER_SIZE];
			if (APR_SUCCESS != (status = apr_file_open(&config, optarg,
					APR_READ, APR_OS_DEFAULT, pmain))) {
				fprintf(stderr, "DBD: Could not open '%s' for read: %s\n",
						optarg, apr_strerror(status, errbuf, MAX_BUFFER_SIZE));
				exit(1);
			}
			if (APR_EOF != (status = apr_file_read_full(config, buffer,
					MAX_BUFFER_SIZE, &len))) {
				if (APR_SUCCESS == status) {
					fprintf(stderr,
							"DBD: '%s' was too long, maximum length %d\n",
							optarg, MAX_BUFFER_SIZE);
					exit(1);
				} else {
					fprintf(stderr, "DBD: Could not read '%s': %s\n", optarg,
							apr_strerror(status, errbuf, MAX_BUFFER_SIZE));
					exit(1);
				}
			}
			dsndbd->params = apr_pstrndup(pmain, buffer, len);
			break;
		}
		case 'f': {
			if (!strcmp(optarg, "-")) {
				status = apr_file_open_stdin(&dsn->mimeparse->in, pmain);
			}
			else if (APR_SUCCESS != (status = apr_file_open(&dsn->mimeparse->in,
					optarg, APR_READ, APR_OS_DEFAULT, pmain))) {
				fprintf(stderr, "DBD: Could not open '%s' for read: %s\n",
						optarg, apr_strerror(status, errbuf, MAX_BUFFER_SIZE));
				exit(1);
			}
			break;
		}
		case 'o': {
			if (!strcmp(optarg, "-")) {
				status = apr_file_open_stdout(&dsn->mimeparse->out, pmain);
			}
			else if (APR_SUCCESS != (status = apr_file_open(&dsn->mimeparse->out,
					optarg, APR_WRITE | APR_CREATE | APR_TRUNCATE,
					APR_OS_DEFAULT, pmain))) {
				fprintf(stderr, "DBD: Could not open '%s' for write: %s\n",
						optarg, apr_strerror(status, errbuf, MAX_BUFFER_SIZE));
				exit(1);
			}
			break;
		}
		}

	}

	/* exit if command args processing failed. */
	if (APR_EOF != status) {
		fprintf(stderr, "DBD: Command line arguments could not be parsed.\n");
		exit(1);
	}

	if (!dsndbd->params) {
		fprintf(
				stderr,
				"DBD: --params or --params-file must be specified, or the " DBD_PARAMS " environment variable must be set.\n");
		exit(1);
	}
	if (!dsndbd->driver_name) {
		fprintf(stderr, "DBD: --driver must be specified.\n");
		exit(1);
	}

	if (dsndbd_open_database(&dsncon, dsndbd, pmain)) {
		exit(1);
	}
	dsn->dsncon = dsncon;

	if ((errno = apr_dbd_transaction_start(dsncon->dsndbd->driver, dsncon->pool,
			dsncon->handle, &dsncon->transaction))) {
		fprintf(stderr, "DBD: Transaction failed to start (using %s): %s",
				dsncon->dsndbd->driver_name, apr_dbd_error(
						dsncon->dsndbd->driver, dsncon->handle, errno));
		exit(1);
	}

	status = read_pipe(dsn->mimeparse);

	if (APR_SUCCESS != status) {
		apr_dbd_transaction_mode_set(dsncon->dsndbd->driver, dsncon->transaction, APR_DBD_TRANSACTION_ROLLBACK);
	}
	else {
		apr_dbd_transaction_mode_set(dsncon->dsndbd->driver, dsncon->transaction, APR_DBD_TRANSACTION_COMMIT);
	}
	apr_dbd_transaction_end(dsncon->dsndbd->driver, dsncon->pool, dsncon->transaction);

	apr_pool_destroy(pmain);

	/* we're done, time to go */
	exit(status != 0);

}
